

#include "navierstokes.h"
#include "navierstokes_bcs.h"
#include "navierstokes_rhs.h"


NavierStokes::NavierStokes(const Settings &settings, Triangulation<dim> &triangulation)
        : triangulation(triangulation)
        , dof_handler_velocity(triangulation)
        , dof_handler_pressure(triangulation)
        , fe_velocity(FE_Q<dim>(settings.fe_degree+1), dim)
        , fe_pressure(settings.fe_degree)
        , quadrature_velocity(settings.fe_degree+2)
        , quadrature_pressure(settings.fe_degree+1)
        , mapping(settings.fe_degree)
        , settings(settings)
        , vel_max_iterations(settings.vel_max_iterations)
        , vel_Krylov_size(settings.vel_Krylov_size)
        , vel_off_diagonals(settings.vel_off_diagonals)
        , vel_eps(settings.vel_eps)
        , vel_diag_strength(settings.vel_diag_strength)
        , imex_pressure_tollerance(settings.iPC_pressure_tollerance)
        , imex_max_pressure_iterations(settings.iPC_max_pressure_iterations)

{
    inlet_velocity_boundary.inlet_velocity_x = settings.inlet_velocity_x;
    inlet_velocity_boundary.inlet_velocity_y = settings.inlet_velocity_y;
    inlet_velocity_boundary.inlet_velocity_z = settings.inlet_velocity_z;
    inlet_velocity_boundary.H = settings.h_channel;
    inlet_velocity_boundary.inlet_type = settings.inlet_type;
    inlet_velocity_boundary.r1 = settings.inner_radius_coaxial_nozzle;
    inlet_velocity_boundary.r2 = settings.outer_radius_coaxial_nozzle;
    inlet_velocity_boundary.channels_angle = settings.channel_angle;
    initial_pressure.l_channel = settings.l_channel;
}



void NavierStokes::setup_system()
{

    // Velocity
    dof_handler_velocity.distribute_dofs(fe_velocity);
    DoFRenumbering::boost::Cuthill_McKee(dof_handler_velocity);


//    FEValuesExtractors::Scalar y_component(1);
    constraints_velocity.clear();
    DoFTools::make_hanging_node_constraints(dof_handler_velocity, constraints_velocity);
//    VectorTools::interpolate_boundary_values(mapping, dof_handler_velocity, 1, Functions::ZeroFunction<dim>(dim), constraints_velocity);
//    VectorTools::interpolate_boundary_values(mapping, dof_handler_velocity, 2, InletVelocity(), constraints_velocity);
//    VectorTools::interpolate_boundary_values(mapping, dof_handler_velocity, 3, Functions::ZeroFunction<dim>(dim), constraints_velocity, fe_velocity.component_mask(y_component));
//    VectorTools::interpolate_boundary_values(mapping, dof_handler_velocity, 4, Functions::ZeroFunction<dim>(dim), constraints_velocity);
    constraints_velocity.close();

    DynamicSparsityPattern dsp(dof_handler_velocity.n_dofs(),dof_handler_velocity.n_dofs());
    DoFTools::make_sparsity_pattern(dof_handler_velocity,
                                    dsp,
                                    constraints_velocity,
                                    false);
    sparsity_pattern_velocity.copy_from(dsp);
    system_matrix.reinit(sparsity_pattern_velocity);

    u_n.reinit(dof_handler_velocity.n_dofs());
    u_n_minus_1.reinit(dof_handler_velocity.n_dofs());
    u_star.reinit(dof_handler_velocity.n_dofs());
    system_rhs.reinit(dof_handler_velocity.n_dofs());
    pressure_gradient_rhs.reinit(dof_handler_velocity.n_dofs());
    vel_Mass.reinit(sparsity_pattern_velocity);
    vel_Laplace.reinit(sparsity_pattern_velocity);
    vel_Advection.reinit(sparsity_pattern_velocity);
    vel_Laplace_plus_Mass.reinit(sparsity_pattern_velocity);
    v_tmp.reinit(dof_handler_velocity.n_dofs());

    u_n_plus_1.reinit(dof_handler_velocity.n_dofs());



    MatrixCreator::create_mass_matrix(dof_handler_velocity,
                                      quadrature_velocity,
                                      vel_Mass,
                                      (const Function<dim> *const) nullptr,
                                      constraints_velocity);

    MatrixCreator::create_laplace_matrix(dof_handler_velocity,
                                         quadrature_velocity,
                                         vel_Laplace,
                                         (const Function<dim> *const) nullptr,
                                         constraints_velocity);


    // Pressure
    dof_handler_pressure.distribute_dofs(fe_pressure);
    DoFRenumbering::boost::Cuthill_McKee(dof_handler_pressure);

    constraints_pressure.clear();
    DoFTools::make_hanging_node_constraints(dof_handler_pressure,
                                            constraints_pressure);
    constraints_pressure.close();

    DynamicSparsityPattern dsp2(dof_handler_pressure.n_dofs(),
                                dof_handler_pressure.n_dofs());
    DoFTools::make_sparsity_pattern(dof_handler_pressure,
                                    dsp2,
                                    constraints_pressure,
                                    false);
    sparsity_pattern_pressure.copy_from(dsp2);
    pres_Laplace.reinit(sparsity_pattern_pressure);
    pres_Mass.reinit(sparsity_pattern_pressure);

    pres_tmp.reinit(dof_handler_pressure.n_dofs());
    pres_n_minus_1.reinit(dof_handler_pressure.n_dofs());
    pres_n.reinit(dof_handler_pressure.n_dofs());
    phi_n_minus_1.reinit(dof_handler_pressure.n_dofs());
    phi_n_plus_1.reinit(dof_handler_pressure.n_dofs());
    phi_n.reinit(dof_handler_pressure.n_dofs());
    pressure_force.reinit(dof_handler_pressure.n_dofs());

    MatrixCreator::create_mass_matrix(dof_handler_pressure,
                                      quadrature_pressure,
                                      pres_Mass,
                                      (const Function<dim> *const) nullptr,
                                      constraints_pressure);


    MatrixCreator::create_laplace_matrix(dof_handler_pressure,
                                         quadrature_pressure,
                                         pres_Laplace,
                                         (const Function<dim> *const) nullptr,
                                         constraints_pressure);


    // Pressure gradient sparsity pattern
    DynamicSparsityPattern dsp_pres_diff(dof_handler_velocity.n_dofs(),
                                dof_handler_pressure.n_dofs());
    DoFTools::make_sparsity_pattern(dof_handler_velocity,
                                    dof_handler_pressure,
                                    dsp_pres_diff);
    sparsity_pattern_pres_vel.copy_from(dsp_pres_diff);
    pres_vel_grad.reinit(sparsity_pattern_pres_vel);

    DynamicSparsityPattern dsp_pres_diff_prova(dof_handler_pressure.n_dofs(), dof_handler_velocity.n_dofs());
    DoFTools::make_sparsity_pattern(dof_handler_pressure, dof_handler_velocity, dsp_pres_diff_prova);
    sparsity_pattern_vel_pres.copy_from(dsp_pres_diff_prova);
    vel_pres_grad.reinit(sparsity_pattern_vel_pres);


    pres_Diff.reinit(sparsity_pattern_pres_vel);

    unsigned int vel_dpc(fe_velocity.dofs_per_cell);
    unsigned int pres_dpc(fe_pressure.dofs_per_cell);

    FullMatrix<double> local_grad(fe_velocity.dofs_per_cell, fe_pressure.dofs_per_cell);

    std::vector<types::global_dof_index> vel_local_dof_indices(fe_velocity.dofs_per_cell);
    std::vector<types::global_dof_index> pres_local_dof_indices(fe_pressure.dofs_per_cell);

    unsigned int nqp(quadrature_velocity.size());

    FEValues<dim> fe_val_vel(fe_velocity, quadrature_velocity,
                             update_values | update_gradients | update_JxW_values);

    FEValues<dim> fe_val_pres(fe_pressure, quadrature_velocity,
                              update_values);


    auto vel_cell = dof_handler_velocity.begin_active();
    const auto endc = dof_handler_velocity.end();
    auto pres_cell = dof_handler_pressure.begin_active();

    for (; vel_cell != endc; ++vel_cell, ++pres_cell) {

        fe_val_vel.reinit(vel_cell);
        fe_val_pres.reinit(pres_cell);

        vel_cell->get_dof_indices(vel_local_dof_indices);
        pres_cell->get_dof_indices(pres_local_dof_indices);

        local_grad = 0;

        for (unsigned int q_point = 0; q_point < nqp; ++q_point) {
            for (unsigned int i = 0; i < vel_dpc; ++i) {
                const unsigned int component_i = fe_velocity.system_to_component_index(i).first;
                for (unsigned int j = 0; j < pres_dpc; ++j) {

                    local_grad(i, j) +=
                            fe_val_pres.shape_value(j, q_point)*
                            fe_val_vel.shape_grad(i, q_point)[component_i] *
                            (-fe_val_vel.JxW(q_point));
                }
            }
        }

        constraints_velocity.distribute_local_to_global(local_grad,
                                                        vel_local_dof_indices,
                                                        constraints_pressure,
                                                        pres_local_dof_indices,
                                                        pres_Diff);
    }


    // Initialize solution vectors
    //VectorTools::interpolate(dof_handler_velocity, Functions::ZeroFunction<dim>(dim), u_n_minus_1);
    //VectorTools::interpolate(dof_handler_velocity, Functions::ZeroFunction<dim>(dim), u_n);
    //VectorTools::interpolate(dof_handler_velocity, Functions::ZeroFunction<dim>(dim), u_n_plus_1);
    VectorTools::interpolate(dof_handler_pressure, initial_pressure, pres_n_minus_1);
    VectorTools::interpolate(dof_handler_pressure, initial_pressure, pres_n);


    VectorTools::interpolate(dof_handler_velocity, initial_velocity, u_n_minus_1);
    VectorTools::interpolate(dof_handler_velocity, initial_velocity, u_n);
    VectorTools::interpolate(dof_handler_velocity, initial_velocity, u_n_plus_1);



    phi_n = 0;
    phi_n_minus_1 = 0;
    phi_n_plus_1 = 0;

    vel_Laplace_plus_Mass = 0.0;
    vel_Laplace_plus_Mass.add(1. / settings.reynolds_number, vel_Laplace);
    vel_Laplace_plus_Mass.add(1.5 / settings.time_step, vel_Mass);
}

