
#ifndef MULTIGRID_NAVIERSTOKES_H
#define MULTIGRID_NAVIERSTOKES_H

#include <deal.II/base/quadrature_lib.h>
#include <deal.II/base/timer.h>
#include <deal.II/base/tensor.h>

#include <deal.II/lac/affine_constraints.h>
#include <deal.II/lac/sparse_matrix.h>
#include <deal.II/lac/dynamic_sparsity_pattern.h>
#include <deal.II/lac/solver_gmres.h>
#include <deal.II/lac/precondition.h>
#include <deal.II/lac/relaxation_block.h>
#include <deal.II/lac/full_matrix.h>
#include <deal.II/lac/sparse_ilu.h>
#include <deal.II/lac/sparse_direct.h>

#include <deal.II/fe/fe_system.h>
#include <deal.II/fe/fe_q.h>
#include <deal.II/fe/mapping_q.h>

#include <deal.II/dofs/dof_accessor.h>
#include <deal.II/dofs/dof_tools.h>
#include <deal.II/dofs/dof_handler.h>
#include <deal.II/dofs/dof_renumbering.h>

#include <deal.II/numerics/vector_tools.h>
#include <deal.II/numerics/matrix_tools.h>
#include <deal.II/numerics/data_out.h>

#include <deal.II/multigrid/mg_constrained_dofs.h>
#include <deal.II/multigrid/multigrid.h>
#include <deal.II/multigrid/mg_transfer.h>
#include <deal.II/multigrid/mg_tools.h>
#include <deal.II/multigrid/mg_coarse.h>
#include <deal.II/multigrid/mg_smoother.h>
#include <deal.II/multigrid/mg_matrix.h>

#include <deal.II/meshworker/mesh_loop.h>
#include <deal.II/base/work_stream.h>


#include "../general.h"
#include "../settings.h"
#include "navierstokes_bcs.h"
#include "navierstokes_rhs.h"

using namespace dealii;

class NavierStokes
{
public:
    NavierStokes(const Settings &settings, Triangulation<dim> &triangulation);

    void setup_system();

    void interpolate_velocity();

    void assemble_system();

    void solve(int step);

    void solve_imex(int step);

    void output_results(DataOut<dim> &data_out) const;

    void compute_errors();

    Triangulation<dim> &triangulation;

    DoFHandler<dim> dof_handler_velocity;
    DoFHandler<dim> dof_handler_pressure;
    FESystem<dim> fe_velocity;
    FE_Q<dim> fe_pressure;
    QGauss<dim> quadrature_velocity;
    QGauss<dim> quadrature_pressure;

    const MappingQ<dim> mapping;

    AffineConstraints<double> constraints_velocity;
    AffineConstraints<double> constraints_pressure;

    SparsityPattern      sparsity_pattern_velocity;
    SparsityPattern      sparsity_pattern_pressure;
    SparsityPattern      sparsity_pattern_pres_vel;
    SparsityPattern sparsity_pattern_vel_pres;

    SparseMatrix<double>    system_matrix;
    Vector<double>          system_rhs;

    SparseILU<double> vel_preconditioner;

    Vector<double> pres_n;
    Vector<double> pres_n_minus_1;
    Vector<double> phi_n;
    Vector<double> phi_n_minus_1;
    Vector<double> phi_n_plus_1;

    SparseMatrix<double> pres_vel_grad;
    SparseMatrix<double> vel_pres_grad;
    SparseMatrix<double> pres_Laplace;
    SparseMatrix<double> pres_Mass;
    SparseMatrix<double> pres_Diff;
    SparseMatrix<double> vel_Laplace;
    SparseMatrix<double> vel_Mass;
    SparseMatrix<double> vel_Advection;
    SparseMatrix<double> vel_Laplace_plus_Mass;

    Vector<double> pres_tmp;
    SparseILU<double> prec_pres_Laplace;
    SparseDirectUMFPACK prec_mass;
    Vector<double> pressure_force;
    Vector<double> pressure_gradient_rhs;
    Vector<double> pres_norm;

    Vector<double> u_n;
    Vector<double> u_n_minus_1;
    Vector<double> u_star;
    Vector<double> v_tmp;
    Vector<double> u_n_plus_1;

    const Settings &settings;

    unsigned int           vel_max_iterations;
    unsigned int           vel_Krylov_size;
    unsigned int           vel_off_diagonals;
    double                 vel_eps;
    double                 vel_diag_strength;

    double                 imex_pressure_tollerance;
    int                    imex_max_pressure_iterations;

    InletVelocity inlet_velocity_boundary;
    InletVelocityLaser inlet_velocity_laser;
    Pressure initial_pressure;
    InitialVelocity initial_velocity;

    double l2_norm_div_u;
};


#endif //MULTIGRID_NAVIERSTOKES_H
