

#include "navierstokes.h"
#include "navierstokes_worker.h"

using IteratorTuple = std::tuple<typename DoFHandler<dim>::active_cell_iterator,
        typename DoFHandler<dim>::active_cell_iterator>;
using IteratorPair = SynchronousIterators<IteratorTuple>;


void assemble_cell_div(const IteratorPair &cell,
                        NS_ScratchData &scratch_data,
                        NS_CopyData &copy_data)
{
    auto vel_cell = std::get<0>(*cell);
    auto pres_cell = std::get<1>(*cell);
    scratch_data.fe_values_velocity.reinit(vel_cell);
    scratch_data.fe_values_pressure_grad.reinit(pres_cell);

    unsigned int nqp(scratch_data.fe_values_velocity.get_quadrature().size());
    unsigned int vel_dpc(scratch_data.fe_values_velocity.get_fe().dofs_per_cell);
    unsigned int pres_dpc(scratch_data.fe_values_pressure_grad.get_fe().dofs_per_cell);

    std::vector<Tensor<2, dim>> present_velocity_gradients(nqp);

    const FEValuesExtractors::Vector velocities(0);

    if(scratch_data.ns.settings.algorithm == "linear")
        scratch_data.fe_values_velocity[velocities].get_function_gradients(scratch_data.ns.u_n, present_velocity_gradients);
    else if (scratch_data.ns.settings.algorithm == "linear_imex")
        scratch_data.fe_values_velocity[velocities].get_function_gradients(scratch_data.ns.u_n_plus_1, present_velocity_gradients);

    copy_data.dof_per_cell_velocity = vel_dpc;
    copy_data.dof_per_cell_pressure = pres_dpc;

    copy_data.cell_rhs.reinit(pres_dpc);
    copy_data.local_dof_indices_velocity.resize(vel_dpc);
    copy_data.local_dof_indices_pressure.resize(pres_dpc);

    vel_cell->get_dof_indices(copy_data.local_dof_indices_velocity);
    pres_cell->get_dof_indices(copy_data.local_dof_indices_pressure);

    copy_data.cell_rhs = 0;

    for (unsigned int q_point = 0; q_point < nqp; ++q_point)
    {

        for (unsigned int i = 0; i < pres_dpc ; ++i) {

            for (unsigned int j = 0; j < vel_dpc; ++j) {

                double present_velocity_divergence = trace(present_velocity_gradients[q_point]);

                copy_data.cell_rhs(i) +=
                        present_velocity_divergence *
                        scratch_data.fe_values_pressure_grad.shape_value(i, q_point)*
                        scratch_data.fe_values_pressure_grad.JxW(q_point);

            }
        }
    }
}


void NavierStokes::compute_errors()
{

    l2_norm_div_u = 0;

    Vector<double> div_u;
    div_u.reinit(dof_handler_pressure.n_dofs());

    auto cell_worker_active = [&](const IteratorPair &cell,
                                  NS_ScratchData &scratch_data,
                                  NS_CopyData &copy_data)
    {
        assemble_cell_div(cell, scratch_data, copy_data);
    };

    auto copier_active = [&](const NS_CopyData &copy_data) {
        constraints_pressure.distribute_local_to_global(copy_data.cell_rhs,
                                                        copy_data.local_dof_indices_pressure,
                                                        div_u);

    };

    WorkStream::run(IteratorPair(IteratorTuple(dof_handler_velocity.begin_active(),
                                               dof_handler_pressure.begin_active())),
                    IteratorPair(IteratorTuple(dof_handler_velocity.end(),
                                               dof_handler_pressure.end())),
                    cell_worker_active,
                    copier_active,
                    NS_ScratchData(*this),
                    NS_CopyData(*this));


    l2_norm_div_u = div_u.l2_norm();

    deallog << "## Continuity l2 norm: " << l2_norm_div_u << std::endl;

}


