

#include "navierstokes.h"
#include <deal.II/lac/matrix_out.h>

void NavierStokes::solve_imex(int step)
{

/** Interpolation Step **/
    statistics->enter_subsection("Navier Stokes Interpolate velocity");
    interpolate_velocity();
    statistics->leave_subsection();
/****/



/** Diffusion Step **/


    assemble_system();

    double error = 1.0;
    int pressure_iterations = 0;

    while(error > imex_pressure_tollerance && pressure_iterations < imex_max_pressure_iterations)
    {
        statistics->enter_subsection("Navier Stokes Diffusion-Assembly");

        pres_tmp.equ(-1.0, pres_n);

        //if(pressure_iterations == 0)
            //pres_tmp.add(-4. / 3., phi_n, 1. / 3., phi_n_minus_1);

        system_rhs = 0.0;
        v_tmp.equ(2.0 / settings.time_step, u_n);
        v_tmp.add(-0.5 / settings.time_step, u_n_minus_1);

        vel_Mass.vmult_add(system_rhs, v_tmp);
        pres_Diff.vmult_add(system_rhs, pres_tmp);

        system_matrix.copy_from(vel_Laplace_plus_Mass);
        system_matrix.add(1.0, vel_Advection);


        //Apply boundary conditions
        std::map<types::global_dof_index, double> boundary_values;
        std::vector<types::boundary_id> boundary_ids;
        boundary_ids = triangulation.get_boundary_ids();
        boundary_values.clear();

        std::set<types::boundary_id> flux_boundary_id;
        flux_boundary_id.insert(settings.outlet_manifold);

        //FEValuesExtractors::Scalar y_component(1);

        std::map<types::boundary_id, const Function<dim>*> function_map;
        auto zero = Functions::ZeroFunction<dim>(dim);
        function_map.insert({settings.outlet_manifold,&zero});


        for (const int &boundary_id : boundary_ids) {

            if (boundary_id == settings.walls_manifold) {
                VectorTools::interpolate_boundary_values(dof_handler_velocity,
                                                         boundary_id,
                                                         Functions::ZeroFunction<dim>(dim),
                                                         boundary_values);
            } else if (boundary_id == settings.inlet_manifold) {
                VectorTools::interpolate_boundary_values(dof_handler_velocity,
                                                         boundary_id,
                                                         inlet_velocity_boundary,
                                                         boundary_values);
            } else if (boundary_id == settings.outlet_manifold) {

                /*
                VectorTools::interpolate_boundary_values(dof_handler_velocity,
                                                             boundary_id,
                                                             Functions::ZeroFunction<dim>(dim),
                                                             boundary_values,
                                                             fe_velocity.component_mask(y_component));
                                                             */


                VectorTools::compute_nonzero_tangential_flux_constraints(dof_handler_velocity,
                                                                         0,
                                                                         flux_boundary_id,
                                                                         function_map,
                                                                         constraints_velocity);


            } else if (boundary_id == settings.laser_manifold) {
                VectorTools::interpolate_boundary_values(dof_handler_velocity,
                                                         boundary_id,
                                                         Functions::ZeroFunction<dim>(dim),
                                                         boundary_values);
            }
        }


        MatrixTools::apply_boundary_values(boundary_values, system_matrix, u_n_plus_1, system_rhs);
        //u_n_minus_1 = u_n;
        statistics->leave_subsection();


        statistics->enter_subsection("Navier Stokes Diffusion-Solve");
        if(((step==2) || (step%settings.reinit_vel_preconditioner==0)) && pressure_iterations == 0)
            vel_preconditioner.initialize(system_matrix, SparseILU<double>::AdditionalData(vel_diag_strength, vel_off_diagonals));

        const double  solve_tolerance = vel_eps * system_rhs.l2_norm();
        SolverControl solver_control(vel_max_iterations, solve_tolerance);

        deallog << "Solving Navier-Stokes with GMRES Standard to tol " << solve_tolerance << std::endl;

        SolverGMRES<> gmres(solver_control, SolverGMRES<>::AdditionalData(vel_Krylov_size));

        Timer time;
        time.start();
        gmres.solve(system_matrix, u_n_plus_1, system_rhs, vel_preconditioner);
        time.stop();

        deallog << " converged in " << solver_control.last_step()
                << " iterations"
                << " in " << time.last_wall_time() << " seconds "
                << "residual: " << solver_control.last_value() << std::endl;

        constraints_velocity.distribute(u_n_plus_1);

        statistics->leave_subsection();

/****/

/** Projection Step **/
        statistics->enter_subsection("Navier Stokes Projection-Assembly");
        pressure_force = 0;
        pres_Diff.Tvmult_add(pressure_force, u_n_plus_1);
        //phi_n_minus_1 = phi_n;

        static std::map<types::global_dof_index , double> bval;

        if(step==2)
        {
            VectorTools::interpolate_boundary_values(dof_handler_pressure, settings.outlet_manifold, Functions::ZeroFunction<dim>(), bval);
        }

        MatrixTools::apply_boundary_values(bval, pres_Laplace, phi_n_plus_1, pressure_force);
        statistics->leave_subsection();

        statistics->enter_subsection("Navier Stokes Projection-Solve");
        const double solve_tolerance_p = vel_eps*pressure_force.l2_norm();

        if(step==2)
        {
            prec_pres_Laplace.initialize(pres_Laplace, SparseILU<double>::AdditionalData(vel_diag_strength, vel_off_diagonals));
        }


        SolverControl solvercontrol(vel_max_iterations, solve_tolerance_p);
        solvercontrol.enable_history_data();

        deallog << "Solving Pressure with CG to tol " << solve_tolerance_p << std::endl;
        SolverCG<> cg(solvercontrol);

        Timer time_p;
        time_p.start();
        cg.solve(pres_Laplace, phi_n_plus_1, pressure_force, prec_pres_Laplace);
        time_p.stop();

        deallog << " converged in " << solvercontrol.last_step()
                << " iterations"
                << " in " << time_p.last_wall_time() << " seconds "
                << "residual: " << solvercontrol.last_value() << std::endl;

        phi_n_plus_1 *= 1.5/settings.time_step;
        constraints_pressure.distribute(phi_n_plus_1);

        statistics->leave_subsection();
/****/

/** Update pressure **/
        statistics->enter_subsection("Navier Stokes Update Pressure");

        pres_n_minus_1 = pres_n;

//     Rotational method implemented
        if((step==2))
            prec_mass.initialize(pres_Mass);
        pres_n = pressure_force;
        prec_mass.solve(pres_n);
        constraints_pressure.distribute(pres_n);
        pres_n.sadd(1.0/settings.reynolds_number, 1.0, pres_n_minus_1);
        pres_n += phi_n_plus_1;

        statistics->leave_subsection();
/****/

        pres_norm = pres_n;
        pres_norm -= pres_n_minus_1;
        error = pres_norm.l2_norm()/(pres_n.l2_norm());

        pressure_iterations++;

        deallog << " Pressure error: " << error << std::endl;
        deallog << " Pressure iterations: " << pressure_iterations << std::endl;

    }

    u_n_minus_1 = u_n;
    u_n = u_n_plus_1;

    phi_n_minus_1 = phi_n;
    phi_n = phi_n_plus_1;

    compute_errors();

}

