#include "cmath"

#include "navierstokes.h"
#include "navierstokes_worker.h"

using IteratorTuple = std::tuple<typename DoFHandler<dim>::active_cell_iterator,
                                 typename DoFHandler<dim>::active_cell_iterator>;
using IteratorPair = SynchronousIterators<IteratorTuple>;


template <class IteratorType>
void assemble_advection_term(const IteratorType &cell,
                             NS_ScratchData &scratch_data,
                             NS_CopyData &copy_data)
{
    const unsigned int dofs_per_cell = scratch_data.fe_values_velocity.get_fe().dofs_per_cell;
    copy_data.dof_per_cell_velocity = dofs_per_cell;
    copy_data.cell_matrix.reinit(dofs_per_cell, dofs_per_cell);

    const unsigned int n_q_points = scratch_data.fe_values_velocity.get_quadrature().size();

    copy_data.local_dof_indices_velocity.resize(dofs_per_cell);
    cell->get_dof_indices(copy_data.local_dof_indices_velocity);

    scratch_data.fe_values_velocity.reinit(cell);

    /** Advection stuff **/
    std::vector<Point<dim>> u_star_local(n_q_points);
    std::vector<Vector<double>> u_star_tmp(n_q_points, Vector<double> (dim));
    std::vector< std::vector< Tensor< 1, dim, double >>> grad_u_star(n_q_points, std::vector< Tensor< 1, dim, double >> (dim));
    std::vector<double> u_star_tmp_divergence(n_q_points);
    scratch_data.fe_values_velocity.get_function_values(scratch_data.ns.u_star, u_star_tmp);

    for (unsigned int d = 0; d < dim; ++d)
    {
        for (unsigned int q = 0; q < n_q_points; ++q)
        {
            u_star_local[q](d) = u_star_tmp[q][d];
        }
    }

    scratch_data.fe_values_velocity.get_function_gradients(scratch_data.ns.u_star, grad_u_star);
    for (unsigned int d = 0; d < dim; ++d)
    {
        for (unsigned int q = 0; q < n_q_points; ++q)
        {
            u_star_tmp[q](d) = 0.;
            u_star_tmp_divergence[q] = 0;
        }
    }
    for (unsigned int d = 0; d < dim; ++d)
    {
        for (unsigned int q = 0; q < n_q_points; ++q)
        {
            u_star_tmp_divergence[q] += grad_u_star[q][d][d];
        }
    }
    /****/


    copy_data.cell_matrix = 0;

    for (unsigned int q_point = 0; q_point < n_q_points; ++q_point)
    {
        for (unsigned int i = 0; i < dofs_per_cell; ++i) {
            const unsigned int component_i = scratch_data.ns.fe_velocity.system_to_component_index(i).first;

            for (unsigned int j = 0; j < dofs_per_cell; ++j) {
                const unsigned int component_j = scratch_data.ns.fe_velocity.system_to_component_index(j).first;

                copy_data.cell_matrix(i, j) +=

                        //Advection
                        ((component_i == component_j) ? 1.0 : 0) *
                        ((u_star_local[q_point] *
                          scratch_data.fe_values_velocity.shape_grad(j, q_point) *
                          scratch_data.fe_values_velocity.shape_value(i, q_point) *
                          scratch_data.fe_values_velocity.JxW(q_point))

                         +

                         (0.5 *
                          u_star_tmp_divergence[q_point] *
                          scratch_data.fe_values_velocity.shape_value(i, q_point) *
                          scratch_data.fe_values_velocity.shape_value(j, q_point) *
                          scratch_data.fe_values_velocity.JxW(q_point)));
            }
        }
    }
}


void NavierStokes::assemble_system()
{
    system_matrix = 0;
    system_rhs = 0;
    vel_Advection = 0;

    auto cell_worker_active = [&](const decltype(dof_handler_velocity.begin_active()) &cell,
                                    NS_ScratchData &scratch_data,
                                    NS_CopyData &copy_data)
        {
            assemble_advection_term(cell, scratch_data, copy_data);
        };


    auto copier_active = [&](const NS_CopyData &copy_data)
        {
            constraints_velocity.distribute_local_to_global(copy_data.cell_matrix,
                                                            copy_data.local_dof_indices_velocity,
                                                            vel_Advection);

        };

    MeshWorker::mesh_loop(dof_handler_velocity.begin_active(),
                          dof_handler_velocity.end(),
                          cell_worker_active,
                          copier_active,
                          NS_ScratchData(*this),
                          NS_CopyData(*this),
                          MeshWorker::assemble_own_cells);

}


void NavierStokes::interpolate_velocity()
{
    u_star.equ(2.0, u_n);
    u_star -= u_n_minus_1;
}









