
// C++:
#include <fstream>
#include <iostream>
#include <random>
#include <cstdio>
#include <ctime>
#include <chrono>
#include <thread>

// Coupled Flows
#include "settings.h"
#include "linearized_ns/navierstokes_worker.h"
#include "linearized_ns/navierstokes_rhs.h"
#include "linearized_ns/navierstokes_bcs.h"
#include "general.h"
#include "linearized_ns/navierstokes.h"
#include "mesher.h"
#include "temperature/temperature.h"
#include "solvers.h"


using namespace dealii;


int main(int argc, char *argv[])
{
  try
    {

      auto start = std::chrono::system_clock::now();
      std::clock_t startcputime = std::clock();

      std::string f((argc > 1) ? (argv[1]) : "../parameter_file/parameter-file.prm");
      Settings settings;
      settings.get_parameters(f);

      std::ofstream log_file;
      if(settings.write_to_file)
      {
        log_file.open(settings.output_directory + settings.output_filename + ".log");
        deallog.attach(log_file);
        deallog.depth_file(1);
        statistics = std::make_unique<TimerOutput>(deallog.get_file_stream(), TimerOutput::summary, TimerOutput::cpu_times);
      }
      else
      {
          deallog.depth_console(1);
          statistics = std::make_unique<TimerOutput>(deallog.get_console(), TimerOutput::summary, TimerOutput::cpu_times);
      }

      std::cout << "Settings file " << f << std::endl;
      deallog << "Settings file " << f << std::endl;

      std::time_t start_time = std::chrono::system_clock::to_time_t(start);
      deallog << "Initial computation at " << std::ctime(&start_time);

      settings.dump();  //print variables

      statistics->enter_subsection("Set Up Data");

      Triangulation<dim> triangulation(Triangulation<dim>::maximum_smoothing);
      mesher mesh(settings);

      std::string salome_extension = "unv";
      std::string gmsh_extension = "msh";
      std::string abaqus_extension = "inp";
      std::string dealii_mesher = "dealii";

      if(std::equal(salome_extension.rbegin(), salome_extension.rend(), settings.mesh_file.rbegin()))
        mesh.load_salome_mesh(triangulation);
      if(std::equal(gmsh_extension.rbegin(), gmsh_extension.rend(), settings.mesh_file.rbegin()))
        mesh.load_gmsh_mesh(triangulation);
      if(std::equal(abaqus_extension.rbegin(), abaqus_extension.rend(), settings.mesh_file.rbegin()))
        mesh.load_abaqus_mesh(triangulation);
      if(std::equal(dealii_mesher.rbegin(), dealii_mesher.rend(), settings.mesh_file.rbegin()))
        mesh.rectangle(triangulation);
        //mesh.channel_with_cylinder(triangulation);



      triangulation.refine_global(settings.refinement_mesh_degree);
      deallog << "3D elements: " << triangulation.n_cells() << std::endl;

      deallog << "Algorithm selected:  " << settings.algorithm << std::endl;

      if(settings.algorithm == "linear" || settings.algorithm == "linear_imex")
      {
        linear(settings, triangulation);
      }

      if(settings.algorithm == "newton" || settings.algorithm == "quasi_newton")
      {
        newton(settings, triangulation);
      }

      auto end = std::chrono::system_clock::now();
      std::chrono::duration<double> elapsed_seconds = end-start;
      std::time_t end_time = std::chrono::system_clock::to_time_t(end);

      deallog << "Finished computation at " << std::ctime(&end_time)
                << " elapsed time: " << elapsed_seconds.count() << "s\n";

      double cpu_duration = (std::clock() - startcputime) / (double)CLOCKS_PER_SEC;
      deallog << "Finished in " << cpu_duration << " seconds [CPU Clock] "
              << " actual cpu time " << (cpu_duration/std::thread::hardware_concurrency()*2) << " seconds"
              << std::endl;

      statistics->print_summary();
    }
  catch (std::exception &exc)
    {
      std::cerr << std::endl
                << std::endl
                << "----------------------------------------------------"
                << std::endl;
      std::cerr << "Exception on processing: " << std::endl
                << exc.what() << std::endl
                << "Aborting!" << std::endl
                << "----------------------------------------------------"
                << std::endl;
      return 1;
    }
  catch (...)
    {
      std::cerr << std::endl
                << std::endl
                << "----------------------------------------------------"
                << std::endl;
      std::cerr << "Unknown exception!" << std::endl
                << "Aborting!" << std::endl
                << "----------------------------------------------------"
                << std::endl;
      return 1;
    }

  return 0;
}
