

#include "newton.h"


NavierStokesNewton::NavierStokesNewton(const Settings &settings, Triangulation<dim> &triangulation)
        : viscosity(1.0 / settings.reynolds_number)
        , gamma(1.0)
        , degree(settings.fe_degree)
        , triangulation(triangulation)
        , fe(FE_Q<dim>(degree + 1), dim, FE_Q<dim>(degree), 1)
        , dof_handler(triangulation)
        , time_step(settings.time_step)
        , settings(settings)

{
    inlet_velocity_boundary.inlet_velocity_x = settings.inlet_velocity_x;
    inlet_velocity_boundary.inlet_velocity_y = settings.inlet_velocity_y;
    inlet_velocity_boundary.inlet_velocity_z = settings.inlet_velocity_z;
    inlet_velocity_boundary.H = settings.h_channel;
    inlet_velocity_boundary.time = 0.0;
    inlet_velocity_boundary.inlet_type = settings.inlet_type;

    setup_dofs();
    initialize_system();
}


void NavierStokesNewton::setup_dofs()
{
    system_matrix.clear();
    pressure_mass_matrix.clear();

    dof_handler.distribute_dofs(fe);

    std::vector<unsigned int> block_component(dim + 1, 0);
    block_component[dim] = 1;
    DoFRenumbering::component_wise(dof_handler, block_component);

    dofs_per_block.resize(2);
    DoFTools::count_dofs_per_block(dof_handler,dofs_per_block,block_component);

    FEValuesExtractors::Vector velocities(0);
    FEValuesExtractors::Scalar y_component(1);

    {
        nonzero_constraints.clear();

        DoFTools::make_hanging_node_constraints(dof_handler, nonzero_constraints);


        std::vector<types::boundary_id> boundary_ids;
        boundary_ids = triangulation.get_boundary_ids();

        for (const int &boundary_id : boundary_ids)
        {

            if (boundary_id == settings.walls_manifold)
            {
                VectorTools::interpolate_boundary_values(dof_handler,
                                                         boundary_id,
                                                         Functions::ZeroFunction<dim>(dim+1),
                                                         nonzero_constraints,
                                                         fe.component_mask(velocities));
            }
            else if (boundary_id == settings.inlet_manifold)
            {
                VectorTools::interpolate_boundary_values(dof_handler,
                                                         boundary_id,
                                                         inlet_velocity_boundary,
                                                         nonzero_constraints,
                                                         fe.component_mask(velocities));
            }
            else if (boundary_id == settings.outlet_manifold)
            {
                VectorTools::interpolate_boundary_values(dof_handler,
                                                         boundary_id,
                                                         Functions::ZeroFunction<dim>(dim+1),
                                                         nonzero_constraints,
                                                         fe.component_mask(y_component));
            }
            else if (boundary_id == settings.laser_manifold)
            {
                VectorTools::interpolate_boundary_values(dof_handler,
                                                         boundary_id,
                                                         Functions::ZeroFunction<dim>(dim+1),
                                                         nonzero_constraints,
                                                         fe.component_mask(velocities));
            }
        }

        nonzero_constraints.close();
    }

    {
        zero_constraints.clear();

        DoFTools::make_hanging_node_constraints(dof_handler, zero_constraints);


        std::vector<types::boundary_id> boundary_ids;
        boundary_ids = triangulation.get_boundary_ids();

        for (const int &boundary_id : boundary_ids)
        {
            if (boundary_id == settings.walls_manifold)
            {
                VectorTools::interpolate_boundary_values(dof_handler,
                                                         boundary_id,
                                                         Functions::ZeroFunction<dim>(dim+1),
                                                         zero_constraints,
                                                         fe.component_mask(velocities));
            }
            else if (boundary_id == settings.inlet_manifold)
            {
                VectorTools::interpolate_boundary_values(dof_handler,
                                                         boundary_id,
                                                         Functions::ZeroFunction<dim>(dim+1),
                                                         zero_constraints,
                                                         fe.component_mask(velocities));
            }
            else if (boundary_id == settings.outlet_manifold)
            {
                VectorTools::interpolate_boundary_values(dof_handler,
                                                         boundary_id,
                                                         Functions::ZeroFunction<dim>(dim+1),
                                                         zero_constraints,
                                                         fe.component_mask(velocities));
            }
            else if (boundary_id == settings.laser_manifold)
            {
                VectorTools::interpolate_boundary_values(dof_handler,
                                                         boundary_id,
                                                         Functions::ZeroFunction<dim>(dim+1),
                                                         zero_constraints,
                                                         fe.component_mask(velocities));
            }
        }

        zero_constraints.close();
    }
}


void NavierStokesNewton::update_boundary_values(double time)
{
    inlet_velocity_boundary.time = time;

    FEValuesExtractors::Vector velocities(0);
    FEValuesExtractors::Scalar y_component(1);
    {
        nonzero_constraints.clear();

        DoFTools::make_hanging_node_constraints(dof_handler, nonzero_constraints);


        std::vector<types::boundary_id> boundary_ids;
        boundary_ids = triangulation.get_boundary_ids();

        for (const int &boundary_id : boundary_ids)
        {
            if (boundary_id == settings.walls_manifold)
            {
                VectorTools::interpolate_boundary_values(dof_handler,
                                                         boundary_id,
                                                         Functions::ZeroFunction<dim>(dim+1),
                                                         nonzero_constraints,
                                                         fe.component_mask(velocities));
            }
            else if (boundary_id == settings.inlet_manifold)
            {
                VectorTools::interpolate_boundary_values(dof_handler,
                                                         boundary_id,
                                                         inlet_velocity_boundary,
                                                         nonzero_constraints,
                                                         fe.component_mask(velocities));
            }
            else if (boundary_id == settings.outlet_manifold)
            {
                VectorTools::interpolate_boundary_values(dof_handler,
                                                         boundary_id,
                                                         Functions::ZeroFunction<dim>(dim+1),
                                                         nonzero_constraints,
                                                         fe.component_mask(y_component));
            }
            else if (boundary_id == settings.laser_manifold)
            {
                VectorTools::interpolate_boundary_values(dof_handler,
                                                         boundary_id,
                                                         Functions::ZeroFunction<dim>(dim+1),
                                                         nonzero_constraints,
                                                         fe.component_mask(velocities));
            }
        }

    }
    nonzero_constraints.close();


}



void NavierStokesNewton::initialize_system()
{
    BlockDynamicSparsityPattern dsp(dofs_per_block, dofs_per_block);
    DoFTools::make_sparsity_pattern(dof_handler, dsp, nonzero_constraints);
    sparsity_pattern.copy_from(dsp);

    system_matrix.reinit(sparsity_pattern);
    present_solution.reinit(dofs_per_block);
    newton_update.reinit(dofs_per_block);
    system_rhs.reinit(dofs_per_block);
    evaluation_point_minus_1.reinit(dofs_per_block);
    present_solution_minus_1.reinit(dofs_per_block);

    deallog << "system matrix:  " << system_matrix.m() << "x" << system_matrix.n()
            << "  system matrix block 1: " << system_matrix.block(0,0).m() << "x" << system_matrix.block(0,0).n()
            << "  system matrix block 0: " << system_matrix.block(1,1).m() << "x" << system_matrix.block(1,1).n()
            << std::endl;
}