#include <tbb/parallel_for_each.h>
#include "particles_dynamic.h"
#include <deal.II/hp/mapping_collection.h>


bool get_line_intersection(double p0_x, double p0_y, double p1_x, double p1_y,
                            double p2_x, double p2_y, double p3_x, double p3_y, Point<dim> &intersection)
{
    double s1_x, s1_y, s2_x, s2_y;
    s1_x = p1_x - p0_x;     s1_y = p1_y - p0_y;
    s2_x = p3_x - p2_x;     s2_y = p3_y - p2_y;

    double s, t;
    s = (-s1_y * (p0_x - p2_x) + s1_x * (p0_y - p2_y)) / (-s2_x * s1_y + s1_x * s2_y);
    t = ( s2_x * (p0_y - p2_y) - s2_y * (p0_x - p2_x)) / (-s2_x * s1_y + s1_x * s2_y);

    if (s >= 0 && s <= 1 && t >= 0 && t <= 1)
    {
        // Collision detected
        intersection(0) = p0_x + (t * s1_x);
        intersection(1) = p0_y + (t * s1_y);
        return true;
    }

    return false; // No collision
}

bool get_surface_intersection(Point<dim> p0, Point<dim> p1,Point<dim> p2,Point<dim> p3,
                              Point<dim> v0,Point<dim> v1, Point<dim> &intersection)
{

    Tensor<1, dim> u = v1-v0;
    double u_norm = u.norm();
    u/=u_norm;

    double t = (-(p0(0)*p1(2)*p2(1)) + p0(0)*p1(1)*p2(2) + p1(2)*p2(1)*v0(0) - p1(1)*p2(2)*v0(0) + p0(0)*p1(2)*v0(1) -
                p1(2)*p2(0)*v0(1) - p0(0)*p2(2)*v0(1) + p1(0)*p2(2)*v0(1) +
                p0(2)*(p1(0)*p2(1) - p2(1)*v0(0) + p1(1)*(-p2(0) + v0(0)) - p1(0)*v0(1) + p2(0)*v0(1)) -
                p0(0)*p1(1)*v0(2) + p1(1)*p2(0)*v0(2) + p0(0)*p2(1)*v0(2) - p1(0)*p2(1)*v0(2) +
                p0(1)*(p1(2)*p2(0) - p1(0)*p2(2) - p1(2)*v0(0) + p2(2)*v0(0) + p1(0)*v0(2) - p2(0)*v0(2)))/
               (-(p1(2)*p2(1)*u[0]) + p1(1)*p2(2)*u[0] - p0(0)*p1(2)*u[1] + p1(2)*p2(0)*u[1] +
                p0(0)*p2(2)*u[1] - p1(0)*p2(2)*u[1] +
                p0(2)*(-(p1(1)*u[0]) + p2(1)*u[0] + p1(0)*u[1] - p2(0)*u[1]) + p0(0)*p1(1)*u[2] -
                p1(1)*p2(0)*u[2] - p0(0)*p2(1)*u[2] + p1(0)*p2(1)*u[2] +
                p0(1)*(p1(2)*u[0] - p2(2)*u[0] - p1(0)*u[2] + p2(0)*u[2]));

    Tensor<1, dim> test_intersection;

    test_intersection[0] = v0(0) + t * u[0];
    test_intersection[1] = v0(1) + t * u[1];
    test_intersection[2] = v0(2) + t * u[2];

    std::vector<Tensor<1, dim>> distance_intersection {test_intersection - p0, test_intersection - p1,
                                                       test_intersection - p2, test_intersection - p3};

    //  p0, p1, p2

    double area1 = cross_product_3d(p0 - p1, p0 - p2).norm()/2.0;
    double alpha1 = cross_product_3d(test_intersection - p1, test_intersection - p2).norm()/(2.0*area1);
    double beta1 = cross_product_3d(test_intersection - p2, test_intersection - p0).norm()/(2.0*area1);
    double gamma1 = 1.0 - alpha1 -beta1;

    bool inside1 = ((0.0 <= alpha1) && (alpha1 <= 1.0)) &&
                   ((0.0 <= beta1) && (beta1 <= 1.0)) &&
                   ((0.0 <= gamma1) && (gamma1 <= 1.0)) ;


    //  p0, p1, p3

    double area2 = cross_product_3d(p0 - p1, p0 - p3).norm()/2.0;
    double alpha2 = cross_product_3d(test_intersection - p1, test_intersection - p3).norm()/(2.0*area2);
    double beta2 = cross_product_3d(test_intersection - p3, test_intersection - p0).norm()/(2.0*area2);
    double gamma2 = 1.0 - alpha2 -beta2;

    bool inside2 = ((0.0 <= alpha2) && (alpha2 <= 1.0)) &&
                   ((0.0 <= beta2) && (beta2 <= 1.0)) &&
                   ((0.0 <= gamma2) && (gamma2 <= 1.0)) ;

    //  p0, p2, p3

    double area3 = cross_product_3d(p0 - p2, p0 - p3).norm()/2.0;
    double alpha3 = cross_product_3d(test_intersection - p2, test_intersection - p3).norm()/(2.0*area3);
    double beta3 = cross_product_3d(test_intersection - p3, test_intersection - p0).norm()/(2.0*area3);
    double gamma3 = 1.0 - alpha3 -beta3;

    bool inside3 = ((0.0 <= alpha3) && (alpha3 <= 1.0)) &&
                   ((0.0 <= beta3) && (beta3 <= 1.0)) &&
                   ((0.0 <= gamma3) && (gamma3 <= 1.0)) ;

    //  p1, p2, p3

    double area4 = cross_product_3d(p1 - p2, p1 - p3).norm()/2.0;
    double alpha4 = cross_product_3d(test_intersection - p2, test_intersection - p3).norm()/(2.0*area4);
    double beta4 = cross_product_3d(test_intersection - p3, test_intersection - p1).norm()/(2.0*area4);
    double gamma4 = 1.0 - alpha4 -beta4;

    bool inside4 = ((0.0 <= alpha4) && (alpha4 <= 1.0)) &&
                   ((0.0 <= beta4) && (beta4 <= 1.0)) &&
                   ((0.0 <= gamma4) && (gamma4 <= 1.0)) ;

    intersection(0) = test_intersection[0];
    intersection(1) = test_intersection[1];
    intersection(2) = test_intersection[2];

    return inside1 || inside2 || inside3 || inside4;
}


void
ParticlesDynamic::advect_particles(const unsigned int step) {


    if(parallelize_particles == false)
    {
        for (auto p = particle_handler.begin(); p != particle_handler.end(); p++)
        {
            advect_particle(p);
        }

        return;
    }


    // mutex used by threads: semaforo per bloccare threads finche tutti non hanno finito
    std::mutex m;

    // condition variable used to see when all threads finish: dice se o no tutti i threads hanno finito
    std::condition_variable cv;

    // number of threads still active: quanti thread ancora sono attivi
    std::atomic<int> counter(0);

    // vector of thread ids and dof ranges
    std::vector<std::thread::id> threadids;
    std::vector<std::pair<std::size_t, std::size_t>> threadranges;


    auto loop = [&](std::size_t begin, std::size_t end) {

        //std::lock_guard<std::mutex> lk(m);

        //std::cout << "thread n:  "  << std::this_thread::get_id() << "  begin:  " << begin
        //          << "  end:  " << end << "  tot: " << next_particle_index << std::endl;

        auto p = particle_handler.begin();
        int i = 0;

        while (i != begin)
        {
            i++;
            p++;
        }

        for(i = begin; i<end; i++, p++)
        {
            advect_particle(p);
        }

    };




    // get number of threads to run: trova numero thread che ci sono
    std::size_t nthreads = std::thread::hardware_concurrency();

    // get number of eqs per thread and the remainder
    // numero particelle per numero di threads
    //physics = numero particelle che abbiamo
    int qq = next_particle_index / nthreads;
    // resto della divisione sopra
    int rr = next_particle_index % nthreads;

    std::size_t s, e;

    // build start/stop ranges per each thread
    for (std::size_t i = 0; i < nthreads; i++) {
        // start
        s = (i == 0) ? 0 : threadranges.back().second;

        // end
        e = (i < rr) ? (s + qq + 1) : (s + qq);

        // store
        threadranges.push_back(std::make_pair(s, e));
    }

    // create threads
    for (std::size_t i = 0; i < nthreads; i++) {
        // increase the number of running threads
        counter++;

        // worker thread
        std::thread worker([&]() {
            // index of thread in the vector of ids
            std::size_t idx;

            // CRITICAL START
            {
                std::lock_guard<std::mutex> lk(m);

                // push the thread id
                threadids.push_back(std::this_thread::get_id());

                // index is the last entry in the vector
                idx = threadids.size() - 1;
            }
            // CRITICAL END

            int start = threadranges[idx].first;
            int stop = threadranges[idx].second;

            // loop over the given equations
            loop(start, stop);

            // CRITICAL START
            {
                std::lock_guard<std::mutex> lk(m);
                counter--;
                cv.notify_all();
            }
            // CRITICAL END
        });

        // il threads parte a fare in run da solo
        worker.detach();
    }

    // wait for all threads
    std::unique_lock<std::mutex> lock(m);
    cv.wait(lock, [&]() { return counter == 0; });
}




void
ParticlesDynamic::advect_particle(Particles::ParticleHandler<dim>::particle_iterator &p)
{

    if(p->get_properties()[type_ignore] > 0 && p->get_properties()[type_block] > 0)
    {
        return;
    }

    Triangulation<dim>::active_cell_iterator current_cell;

    try
    {
        current_cell = GridTools::find_active_cell_around_point(triangulation, p->get_location());
    }
    catch (...)
    {
        return;
    }

    Point<dim> ref_p = StaticMappingQ1<dim>::mapping.transform_real_to_unit_cell(current_cell, p->get_location());
    p->set_reference_location(ref_p);


    /** Updating temperature field **/

    double temperature_interpolated_field;
    double laser_field;

        /** Temperature field **/

    std::vector<types::global_dof_index> temp_dof_indices(temperature_fe.n_dofs_per_cell());
    const auto &temp_dh_cell = typename DoFHandler<dim, dim>::cell_iterator(*current_cell, &dof_handler_temperature);

    temp_dh_cell->get_dof_indices(temp_dof_indices);
    temperature_interpolated_field = 0;

    for (unsigned int k = 0; k < temperature_fe.n_dofs_per_cell(); ++k) {
        temperature_interpolated_field +=
                temperature_fe.shape_value(k, p->get_reference_location()) *
                temperature_vector(temp_dof_indices[k]);
    }

        /** Laser field **/

    std::vector<types::global_dof_index> laser_dof_indices(laser_fe.n_dofs_per_cell());
    const auto &laser_dh_cell = typename DoFHandler<dim, dim>::cell_iterator(*current_cell, &dof_handler_laser);

    laser_dh_cell->get_dof_indices(laser_dof_indices);
    laser_field = 0;

    for (unsigned int k = 0; k < laser_fe.n_dofs_per_cell(); ++k) {

        laser_field += laser_fe.shape_value(k, p->get_reference_location())
                       * laser_vector(laser_dof_indices[k]);
    }

        /** Particle temperature updated **/


    double particle_area;
    double particle_volume;

    if(dim ==2)
    {
        particle_area = 2*(M_PI * (p->get_properties()[num_diameter] / 2));
        particle_volume = (M_PI * std::pow(p->get_properties()[num_diameter] / 2, 2));
    }
    else
    {
        particle_area = 4*(M_PI * std::pow(p->get_properties()[num_diameter] / 2, 2));
        particle_volume = (4.0 / 3.0) * (M_PI * std::pow(p->get_properties()[num_diameter] / 2, 3));
    }

    double particle_mass = particle_density * particle_volume;

    auto laser_power_absorption =  particle_absorptivity_rate * laser_field * particle_area;

    auto particle_fluid_convection = particle_convection_coefficient *
                                     (p->get_properties()[num_temperature] - temperature_interpolated_field) *
                                     particle_area;

    double delta = 0;

    if(p->get_properties()[num_temperature] > settings.solid_temperature &&
       p->get_properties()[num_temperature] < settings.liquid_temperature)
    {
        delta = 1;
    }

    auto coeff_A = particle_mass*
                   (particle_specific_heat +
                    (delta * settings.latent_heat/(settings.liquid_temperature - settings.solid_temperature)));

    auto new_particle_temperature =
            p->get_properties()[num_temperature] + ((dt / coeff_A) *
                                                    (laser_power_absorption - particle_fluid_convection));

    double liquid_fraction = 0;
    if(new_particle_temperature > settings.solid_temperature /* && new_particle_temperature < settings.liquid_temperature*/ )
    {
        liquid_fraction = ((new_particle_temperature - settings.solid_temperature)/
                           (settings.liquid_temperature - settings.solid_temperature));
    }

    double new_liquid_fraction = p->get_properties()[num_liquid_fraction] + liquid_fraction;
    if(new_liquid_fraction > 1)
        new_liquid_fraction = 1;


    std::vector<double>  update_temperature_properties(num_properties);
    update_temperature_properties[num_temperature] = new_particle_temperature;
    update_temperature_properties[num_fluid_temperature] = temperature_interpolated_field;
    update_temperature_properties[num_velocity_x] = p->get_properties()[num_velocity_x];
    update_temperature_properties[num_fluid_velocity_x] = p->get_properties()[num_fluid_velocity_x];
    update_temperature_properties[num_velocity_y] = p->get_properties()[num_velocity_y];
    update_temperature_properties[num_fluid_velocity_y] = p->get_properties()[num_fluid_velocity_y];
    update_temperature_properties[num_velocity_z] = p->get_properties()[num_velocity_z];
    update_temperature_properties[num_fluid_velocity_z] = p->get_properties()[num_fluid_velocity_z];
    update_temperature_properties[num_laser_power] = laser_field;
    update_temperature_properties[num_liquid_fraction] = new_liquid_fraction;
    update_temperature_properties[num_diameter] = p->get_properties()[num_diameter];
    update_temperature_properties[type_ignore] = p->get_properties()[type_ignore];
    update_temperature_properties[type_advect] = p->get_properties()[type_advect];
    update_temperature_properties[type_block] = p->get_properties()[type_block];

    p->set_properties(update_temperature_properties);


    /** CHECK IF BLOCK PARTICLE **/
/*
    if(p->get_properties()[type_block] > 0)
        return;
        */


    /** Updating velocity field **/

    const ComponentMask vel_comps = ComponentMask(fe_velocity.n_components(), true);

    std::vector<unsigned int> space_gtl(fe_velocity.n_components(), numbers::invalid_unsigned_int);
    for (unsigned int i = 0, j = 0; i < space_gtl.size(); ++i)
        if (vel_comps[i])
            space_gtl[i] = j++;

    std::vector<double> velocity_interpolated_field;
    velocity_interpolated_field.resize(dim);

    std::vector<types::global_dof_index> vel_dof_indices(fe_velocity.n_dofs_per_cell());
    const auto &vel_dh_cell = typename DoFHandler<dim, dim>::cell_iterator(*current_cell, &dof_handler_velocity);
    vel_dh_cell->get_dof_indices(vel_dof_indices);
    velocity_interpolated_field.assign(dim, 0);

    for (unsigned int j = 0; j < fe_velocity.n_dofs_per_cell(); ++j) {
        const auto comp_j = space_gtl[fe_velocity.system_to_component_index(j).first];


        if (comp_j != numbers::invalid_unsigned_int)
            velocity_interpolated_field[comp_j] +=
                    fe_velocity.shape_value(j, p->get_reference_location()) *
                    u_n(vel_dof_indices[j]);

    }

        /** Move particles **/

    Point<dim> loc = p->get_location();

    Tensor<1, dim> old_particle_velocity;
    old_particle_velocity[0] = p->get_properties()[num_velocity_x];
    old_particle_velocity[1] = p->get_properties()[num_velocity_y];

    int fluid_x_sign = (signbit(velocity_interpolated_field[0]) ? -1 : 1);
    int fluid_y_sign = (signbit(velocity_interpolated_field[1]) ? -1 : 1);
    int par_x_sign = (signbit(old_particle_velocity[0]) ? -1 : 1);
    int par_y_sign = (signbit(old_particle_velocity[1]) ? -1 : 1);

    Tensor<1, dim> relative_velocity;

    relative_velocity[0] = fluid_x_sign * std::pow(velocity_interpolated_field[0], 2) -
                           (par_x_sign) * std::pow(old_particle_velocity[0], 2);
    relative_velocity[1] = fluid_y_sign * std::pow(velocity_interpolated_field[1], 2) -
                           (par_y_sign) * std::pow(old_particle_velocity[1], 2);
    if (dim > 2) {
        old_particle_velocity[2] = p->get_properties()[num_velocity_z];
        int fluid_z_sign = (signbit(velocity_interpolated_field[2]) ? -1 : 1);
        int par_z_sign = (signbit(old_particle_velocity[2]) ? -1 : 1);
        relative_velocity[2] = fluid_z_sign * std::pow(velocity_interpolated_field[2], 2) -
                               (par_z_sign) * std::pow(old_particle_velocity[2], 2);
    }

        /** Particle velocity updated **/

    double drag_coefficient;
/*
    double particle_reynolds = p->get_properties()[num_diameter] * (1/settings.reynolds_number) * relative_velocity.norm();
    if(particle_reynolds <= 1.0)
        drag_coefficient = 24.0 / particle_reynolds;
    else if(particle_reynolds > 1.0 && particle_reynolds <= 1000.0)
        drag_coefficient = 24.0 / particle_reynolds * (1+ 0.15 * pow(particle_reynolds, 0.687));
    else
        drag_coefficient = drag_coefficient_high_reynolds;
        */

    drag_coefficient = drag_coefficient_high_reynolds;

    //std::cout << "drag:  " << drag_coefficient << std::endl;


    auto drag_force = 0.5 * drag_coefficient * fluid_density * particle_area * relative_velocity;
    auto buoyancy = vector_gravity * particle_density * particle_volume *
                    ((particle_density - fluid_density) / particle_density);



    auto new_particle_velocity = old_particle_velocity + (dt / particle_mass) * (drag_force + buoyancy);

    /** Check particle new position **/

    Point<dim> newLoc = loc + dt * new_particle_velocity;

    try {

        auto coming_cell = GridTools::find_active_cell_around_point(triangulation, newLoc);
        p->set_location(newLoc);

        std::vector<double>  update_velocity_properties(num_properties);
        update_velocity_properties[num_temperature] = p->get_properties()[num_temperature];
        update_velocity_properties[num_fluid_temperature] = p->get_properties()[num_fluid_temperature];
        update_velocity_properties[num_velocity_x] = new_particle_velocity[0];
        update_velocity_properties[num_fluid_velocity_x] = velocity_interpolated_field[0];
        update_velocity_properties[num_velocity_y] = new_particle_velocity[1];
        update_velocity_properties[num_fluid_velocity_y] = velocity_interpolated_field[1];
        update_velocity_properties[num_velocity_z] = p->get_properties()[num_velocity_z];
        update_velocity_properties[num_fluid_velocity_z] = p->get_properties()[num_fluid_velocity_z];
        update_velocity_properties[num_laser_power] = p->get_properties()[num_laser_power];
        update_velocity_properties[num_liquid_fraction] = p->get_properties()[num_liquid_fraction];
        update_velocity_properties[num_diameter] = p->get_properties()[num_diameter];
        update_velocity_properties[type_ignore] = p->get_properties()[type_ignore];
        update_velocity_properties[type_advect] = p->get_properties()[type_advect];
        update_velocity_properties[type_block] = p->get_properties()[type_block];

        if (dim > 2) {
            update_velocity_properties[num_velocity_z] = new_particle_velocity[2];
            update_velocity_properties[num_fluid_velocity_z] = velocity_interpolated_field[2];
        }

        p->set_properties(update_velocity_properties);


    }
    catch (...)
    {
        double distance = std::numeric_limits<double>::max();
        Point<dim> intersection;
        std::pair<unsigned int, unsigned int> rebound_cell;
        unsigned int rebound_face;

        for (auto &wc:wall_cells) {
            Triangulation<dim>::active_cell_iterator wall_cell_iterator(&triangulation, wc.first, wc.second);

            for (unsigned int wcf = 0; wcf < GeometryInfo<dim>::faces_per_cell; ++wcf) {
                bool on_boundary = wall_cell_iterator->face(wcf)->at_boundary();

                if (on_boundary) {
                    Point<dim> new_intersection;
                    bool intersecated = false;

                    if (dim == 2) {
                        intersecated = get_line_intersection(loc(0), loc(1), newLoc(0), newLoc(1),
                                                             wall_cell_iterator->face(wcf)->vertex(0)(0),
                                                             wall_cell_iterator->face(wcf)->vertex(0)(1),
                                                             wall_cell_iterator->face(wcf)->vertex(1)(0),
                                                             wall_cell_iterator->face(wcf)->vertex(1)(1),
                                                             new_intersection);
                    } else {
                        intersecated = get_surface_intersection(wall_cell_iterator->face(wcf)->vertex(0),
                                                                wall_cell_iterator->face(wcf)->vertex(1),
                                                                wall_cell_iterator->face(wcf)->vertex(2),
                                                                wall_cell_iterator->face(wcf)->vertex(3),
                                                                loc, newLoc, new_intersection);
                    }

                    if (intersecated == true) {
                        double new_distance = loc.distance(new_intersection);
                        if (new_distance < distance) {
                            distance = new_distance;
                            intersection = new_intersection;
                            rebound_cell = wc;
                            rebound_face = wcf;
                        }
                    }
                }
            }
        }

        Triangulation<dim>::active_cell_iterator rebound_cell_iterator(&triangulation, rebound_cell.first,
                                                                       rebound_cell.second);

        auto rebound_boundary_id = rebound_cell_iterator->face(rebound_face)->boundary_id();

        if (rebound_boundary_id == settings.walls_manifold)
        {
            FlatManifold<dim, dim> manifold;

            Point<dim> vertex1 = rebound_cell_iterator->face(rebound_face)->vertex(0);
            Point<dim> vertex2 = rebound_cell_iterator->face(rebound_face)->vertex(1);
            Point<dim> vertex3;

            if (dim > 2)
                vertex3 = rebound_cell_iterator->face(rebound_face)->vertex(2);

            const Tensor<1, dim> tangent1 = manifold.get_tangent_vector(vertex1, vertex2);

            Tensor<1, dim> tangent2;
            if (dim > 2)
                tangent2 = manifold.get_tangent_vector(vertex2, vertex3);

            // then rotate it by 90 degrees

            Tensor<1, dim> normal;

            if (dim == 2) {
                normal = cross_product_2d(tangent1);
            } else {
                normal = cross_product_3d(tangent1, tangent2);
            }

            auto boundary_normal = (normal / normal.norm());

            auto v = new_particle_velocity;

            auto u = scalar_product(v, boundary_normal) * (boundary_normal);

            auto w = v - u;

            auto rebound_velocity = w - u;

            const Point<dim> rebound = loc + dt * (w);

            p->set_location(rebound);

            new_particle_velocity = rebound_velocity;

            std::vector<double>  update_rebound_properties(num_properties);
            update_rebound_properties[num_temperature] = p->get_properties()[num_temperature];
            update_rebound_properties[num_fluid_temperature] = p->get_properties()[num_fluid_temperature];
            update_rebound_properties[num_velocity_x] = new_particle_velocity[0];
            update_rebound_properties[num_fluid_velocity_x] = p->get_properties()[num_fluid_velocity_x];
            update_rebound_properties[num_velocity_y] = new_particle_velocity[1];
            update_rebound_properties[num_fluid_velocity_y] = p->get_properties()[num_fluid_velocity_y];;
            update_rebound_properties[num_velocity_z] = p->get_properties()[num_velocity_z];
            update_rebound_properties[num_fluid_velocity_z] = p->get_properties()[num_fluid_velocity_z];
            update_rebound_properties[num_laser_power] = p->get_properties()[num_laser_power];
            update_rebound_properties[num_liquid_fraction] = p->get_properties()[num_liquid_fraction];
            update_rebound_properties[num_diameter] = p->get_properties()[num_diameter];
            update_rebound_properties[type_ignore] = p->get_properties()[type_ignore];
            update_rebound_properties[type_advect] = p->get_properties()[type_advect];
            update_rebound_properties[type_block] = p->get_properties()[type_block];

            if (dim > 2) {
                update_rebound_properties[num_velocity_z] = new_particle_velocity[2];
            }

            p->set_properties(update_rebound_properties);
        }
        else if (rebound_boundary_id == settings.substrate_manifold)
        {
            if (block_particle)
            {
                Point<dim> block_position;
                block_position(0) = loc(0);
                block_position(1) = loc(1);
                block_position(2) = settings.substrate_location + 1e-10;

                p->set_location(block_position);


                std::vector<double>  update_block_properties(num_properties);
                update_block_properties[num_temperature] = p->get_properties()[num_temperature];
                update_block_properties[num_fluid_temperature] = p->get_properties()[num_fluid_temperature];
                update_block_properties[num_velocity_x] = 0.0;
                update_block_properties[num_fluid_velocity_x] = p->get_properties()[num_fluid_velocity_x];
                update_block_properties[num_velocity_y] = 0.0;
                update_block_properties[num_fluid_velocity_y] = p->get_properties()[num_fluid_velocity_y];;
                update_block_properties[num_velocity_z] = p->get_properties()[num_velocity_z];
                update_block_properties[num_fluid_velocity_z] = p->get_properties()[num_fluid_velocity_z];
                update_block_properties[num_laser_power] = p->get_properties()[num_laser_power];
                update_block_properties[num_liquid_fraction] = p->get_properties()[num_liquid_fraction];
                update_block_properties[num_diameter] = p->get_properties()[num_diameter];
                update_block_properties[type_ignore] = p->get_properties()[type_ignore];
                update_block_properties[type_advect] = p->get_properties()[type_advect];
                update_block_properties[type_block] = 1.0;

                if (dim > 2) {
                    update_block_properties[num_velocity_z] = 0.0;
                }

                p->set_properties(update_block_properties);
            }
            else
            {
                std::vector<double>  update_outlet_properties(num_properties);
                update_outlet_properties[num_temperature] = p->get_properties()[num_temperature];
                update_outlet_properties[num_fluid_temperature] = p->get_properties()[num_fluid_temperature];
                update_outlet_properties[num_velocity_x] = p->get_properties()[num_velocity_x];
                update_outlet_properties[num_fluid_velocity_x] = p->get_properties()[num_fluid_velocity_x];
                update_outlet_properties[num_velocity_y] = p->get_properties()[num_velocity_y];
                update_outlet_properties[num_fluid_velocity_y] = p->get_properties()[num_fluid_velocity_y];;
                update_outlet_properties[num_velocity_z] = p->get_properties()[num_velocity_z];
                update_outlet_properties[num_fluid_velocity_z] = p->get_properties()[num_fluid_velocity_z];
                update_outlet_properties[num_laser_power] = p->get_properties()[num_laser_power];
                update_outlet_properties[num_liquid_fraction] = p->get_properties()[num_liquid_fraction];
                update_outlet_properties[num_diameter] = p->get_properties()[num_diameter];
                update_outlet_properties[type_ignore] = 1.0;
                update_outlet_properties[type_advect] = p->get_properties()[type_advect];
                update_outlet_properties[type_block] = p->get_properties()[type_block];

                p->set_properties(update_outlet_properties);
            }

        }
        else
        {
            std::vector<double>  outlet_properties(num_properties);
            outlet_properties[num_temperature] = p->get_properties()[num_temperature];
            outlet_properties[num_fluid_temperature] = p->get_properties()[num_fluid_temperature];
            outlet_properties[num_velocity_x] = p->get_properties()[num_velocity_x];
            outlet_properties[num_fluid_velocity_x] = p->get_properties()[num_fluid_velocity_x];
            outlet_properties[num_velocity_y] = p->get_properties()[num_velocity_y];
            outlet_properties[num_fluid_velocity_y] = p->get_properties()[num_fluid_velocity_y];;
            outlet_properties[num_velocity_z] = p->get_properties()[num_velocity_z];
            outlet_properties[num_fluid_velocity_z] = p->get_properties()[num_fluid_velocity_z];
            outlet_properties[num_laser_power] = p->get_properties()[num_laser_power];
            outlet_properties[num_liquid_fraction] = p->get_properties()[num_liquid_fraction];
            outlet_properties[num_diameter] = p->get_properties()[num_diameter];
            outlet_properties[type_ignore] = 1.0;
            outlet_properties[type_advect] = p->get_properties()[type_advect];
            outlet_properties[type_block] = p->get_properties()[type_block];

            p->set_properties(outlet_properties);

        }
    }
}
