
#include "particles_dynamic.h"
#include "../linearized_ns/navierstokes.h"
#include "../general.h"

using namespace dealii;

ParticlesDynamic::ParticlesDynamic(const Settings &settings, const NavierStokes &ns, const Temperature &temperature, const Laser &laser,Triangulation<dim> &triangulation)
        :
        triangulation(triangulation),
        settings(settings),
        u_n(ns.u_n),
        fe_velocity(ns.fe_velocity),
        dof_handler_velocity(ns.dof_handler_velocity),
        dof_handler_temperature(temperature.dof_handler_temperature),
        inlet_particle_mass(settings.inlet_particle_mass),
        inlet_duration(settings.inlet_duration),
        particle_density(settings.particle_density),
        drag_coefficient_high_reynolds(settings.drag_coefficient),
        particle_diameter(settings.particle_diameter),
        fluid_density(settings.fluid_density),
        gravity(settings.gravity),
        dt(settings.time_step),
        laser_vector(laser.laser_radiation),
        laser_fe(laser.fe_laser),
        dof_handler_laser(laser.dof_handler_laser),
        temperature_vector(temperature.temperature),
        temperature_fe(temperature.fe_temperature),
        random_device{},
        generate_random_numbers{random_device()},
        gaussian_distribution{settings.particle_diameter, settings.particle_variance}

{
    block_particle = settings.block_particles_on_boundary;
    vector_gravity[dim - 1] = -gravity;
    initial_particle_velocity[0] = settings.initial_particle_velocity_x;
    initial_particle_velocity[1] = settings.initial_particle_velocity_y;
    if(dim > 2)
        initial_particle_velocity[2] = settings.initial_particle_velocity_z;

    initial_particle_temperature = settings.initial_particle_temperature;
    particle_specific_heat = settings.particle_specific_heat;
    particle_absorptivity_rate = settings.particle_absorptivity_rate;
    particle_convection_coefficient = settings.particle_convection_coefficient;

    parallelize_particles = settings.parallelize_particles;

    inlet_slots = (int)ceil(inlet_duration /dt);
    total_mass_to_inlet = inlet_particle_mass * inlet_duration;  //[Kg]
    mass_per_slot = total_mass_to_inlet/inlet_slots;             //[Kg]

    double particle_mean_volume = (4.0 / 3.0) * (M_PI * std::pow(particle_diameter / 2, 3));
    particle_mean_mass = particle_density * particle_mean_volume;          //[Kg]
    n_particles_per_slot = (int)ceil(mass_per_slot/particle_mean_mass);

    next_particle_index = 0;
    next_slot_index = 0;

    inlet_particle_boundary_id_1 = settings.inlet_manifold;

    for (auto &cell : triangulation.active_cell_iterators())
    {
        for (unsigned int face = 0; face < GeometryInfo<dim>::faces_per_cell;
             ++face)
        {
            auto current_boundary_id = cell->face(face)->boundary_id();

            if (current_boundary_id == inlet_particle_boundary_id_1)
            {
                const auto cell_lev = cell->level();
                const auto cell_idx = cell->index();

                inlet_cells.push_back(std::make_pair(cell_lev, cell_idx));
            }
            if(current_boundary_id == settings.walls_manifold ||
               current_boundary_id == settings.outlet_manifold ||
               current_boundary_id == settings.substrate_manifold)
            {
                const auto cell_lev = cell->level();
                const auto cell_idx = cell->index();

                wall_cells.push_back(std::make_pair(cell_lev, cell_idx));
            }
        }
    }

    /*
    n_particles_per_cell_per_slot = n_particles_per_slot/inlet_cells.size();
    if(n_particles_per_cell_per_slot < 1)
        n_particles_per_cell_per_slot = 1;
        */


    //Particle_handler
    particle_handler.initialize(triangulation, StaticMappingQ1<dim>::mapping, num_properties);


    deallog << std::endl;
    deallog << "PARTICLES PARAMETERS " << std::endl;
    deallog << "Inlet particles velocity_x: " << settings.initial_particle_velocity_x << std::endl;
    deallog << "Inlet particles velocity_y: " << settings.initial_particle_velocity_y << std::endl;
    deallog << "Inlet particles velocity_z: " << settings.initial_particle_velocity_z << std::endl;
    deallog << "Initial particles temperature: " << settings.initial_particle_temperature << " [K]" << std::endl;
    deallog << "Particles flow rate: " << settings.inlet_particle_mass << std::endl;
    deallog << "Particle density: " << particle_density << " [Kg/m^3]" << std::endl;
    deallog << "Particle mean diameter: " << particle_diameter * 1000 << " [mm]" << std::endl;
    deallog << "Particle inlet duration: " << settings.inlet_duration << " [s]" << std::endl;
    deallog << "Drag coefficient: "  << settings.drag_coefficient <<  std::endl;
    deallog << "Fluid Density: " << fluid_density << " [Kg/m^3]" << std::endl;
    deallog << "Particle specific heat: " << settings.particle_specific_heat << " [J/(K Kg)]" << std::endl;
    deallog << "Particle absorptivity rate: " << settings.particle_absorptivity_rate << " [%]" << std::endl;
    deallog << "Particle convection coefficient: " << settings.particle_convection_coefficient << " [W/(m^2 K)]" << std::endl;
    deallog << "Latent heat: " << settings.latent_heat << " [J/kg]" << std::endl;
    deallog << "Liquid temperature  : " << settings.liquid_temperature << " [K]" << std::endl;
    deallog << "Solid temperature: " << settings.solid_temperature << " [K]" << std::endl;
    deallog << "Total Particles: " << n_particles_per_slot*inlet_slots << std::endl;
    deallog << std::endl;
    deallog << "LASER PARAMETERS " << std::endl;
    deallog << "Laser power: " << settings.laser_power << " [W]" << std::endl;
    deallog << "Laser radius: " << settings.laser_radius << " [m]" << std::endl;
    deallog << "Laser focal plane: " << settings.laser_focal_plane << " [m]" << std::endl;
    deallog << "Laser angle: " << settings.laser_angle << " [°]" << std::endl;
    deallog << std::endl;
}


