//
// Created by Edoardo Costantini on 14/12/2019.
//

#ifndef PARTICLES_DYNAMIC_H
#define PARTICLES_DYNAMIC_H

#include <deal.II/numerics/solution_transfer.h>

#include <deal.II/fe/mapping.h>

#include <deal.II/particles/particle.h>
#include <deal.II/particles/particle_handler.h>
#include <deal.II/particles/particle_iterator.h>
#include <deal.II/particles/particle_accessor.h>
#include <deal.II/particles/generators.h>
#include <deal.II/particles/utilities.h>

#include "../settings.h"
#include "../linearized_ns/navierstokes.h"
#include "../temperature/temperature.h"
#include "../laser/laser.h"
#include "../general.h"

using namespace dealii;

class NavierStokes;
class Temperature;
class Laser;


class ParticleHasher
{
public:
    std::size_t operator()(const Particles::Particle<dim>* p) const
    {
        std::hash<std::size_t> h;
        return h(p->get_id());
    }
};

class ParticleCompare
{
public:

    bool operator()(const Particles::Particle<dim>* p1, const Particles::Particle<dim>* p2) const
    {
        return p1->get_id() == p2->get_id();
    }
};



class ParticlesDynamic {

public:
    ParticlesDynamic(const Settings &settings, const NavierStokes &ns, const Temperature &temperature, const Laser &laser, Triangulation<dim> &triangulation);
    void generate_particles();
    void advect_particles(const unsigned int step);
    void advect_particle(Particles::ParticleHandler<dim>::particle_iterator &p);
    void output_results(const unsigned int step);
    void generate_particles_on_boundary(unsigned int particles_to_generate);
    void generate_particles_on_ring(unsigned int n_particles_per_slot, double r1, double r2);
    void generate_one_particle(Point<dim> &p);

    enum properties_numbering
    {
        num_temperature = 0,
        num_fluid_temperature,
        num_velocity_x,
        num_fluid_velocity_x,
        num_velocity_y,
        num_fluid_velocity_y,
        num_velocity_z,
        num_fluid_velocity_z,
        num_laser_power,
        num_liquid_fraction,
        num_diameter,
        type_ignore,
        type_advect,
        type_block,
        num_properties
    };

private:
    const Triangulation<dim> &triangulation;
    const Settings &settings;

    /****from navierstokes ****/
    const Vector<double> &u_n;
    const FESystem<dim> &fe_velocity;
    const DoFHandler<dim>  &dof_handler_velocity;

    /**************************/

    /**** from laser ****/
    const Vector<double> &laser_vector;
    const FE_Q<dim>  &laser_fe;
    const DoFHandler<dim>  &dof_handler_laser;
    /**************************/

    /**** from temperature ****/
    const FE_Q<dim>  &temperature_fe;
    const Vector<double> &temperature_vector;
    const DoFHandler<dim>  &dof_handler_temperature;
    /**************************/

    /** particles coefficients  **/
    double inlet_particle_mass;
    double inlet_duration;
    double particle_density;
    double drag_coefficient_high_reynolds;
    double particle_diameter;
    double fluid_density;
    double particle_specific_heat;
    double particle_absorptivity_rate;
    double particle_convection_coefficient;
    const double gravity;
    Tensor<1,dim> vector_gravity;
    const double dt;


    int inlet_slots;
    double total_mass_to_inlet;
    double mass_per_slot;
    double particle_mean_mass;
    int n_particles_per_slot;
    unsigned long n_particles_per_cell_per_slot;

    /** particles structures *****/
    //std::multimap<Particles::internal::LevelInd , Particles::Particle<dim>> particles;
    std::unordered_set<Particles::Particle<dim>*, ParticleHasher, ParticleCompare> particles;
    unsigned int next_particle_index;
    unsigned int next_slot_index;
    int inlet_particle_boundary_id_1;
    Tensor<1, dim> initial_particle_velocity;
    double initial_particle_temperature;
    std::vector<std::pair<unsigned int, unsigned int>> inlet_cells;
    std::vector<std::pair<unsigned int, unsigned int>> wall_cells;
    std::vector<unsigned int> particles_to_remove;
    Particles::ParticleHandler<dim,dim> particle_handler;
    bool block_particle;
    /*****************************/

    bool parallelize_particles;

    std::random_device random_device;
    std::mt19937 generate_random_numbers;
    std::normal_distribution<> gaussian_distribution;

};


#endif //PARTICLES_DYNAMIC_H
