#include "solvers.h"


void linear(const Settings &settings, Triangulation<dim> &triangulation)
{
    std::ofstream csv_info(settings.output_directory + "linear_navierstokes.csv");

    NavierStokes navierstokes(settings, triangulation);
    navierstokes.setup_system();

    Temperature temperature(settings, triangulation);
    temperature.setup_system();

    Laser laser(settings, triangulation);
    laser.setup_system();

    deallog << "Number of velocity degrees of freedom: "
            << navierstokes.dof_handler_velocity.n_dofs() << std::endl
            << "Number of pressure degrees of freedom: "
            << navierstokes.dof_handler_pressure.n_dofs() << std::endl
            << "Number of temperature degrees of freedom: "
            << temperature.dof_handler_temperature.n_dofs() << std::endl;


    ParticlesDynamic particles_dynamic(settings,navierstokes,temperature, laser, triangulation);
    particles_dynamic.generate_particles();

    statistics->leave_subsection();

    int n_step = static_cast<int>(settings.simulation_time / settings.time_step);
    int step = 2;
    for (double time = settings.time_step; time <= settings.simulation_time; time += settings.time_step) {
        deallog << "Time: " << time << " step/tot: " << step << "/" << n_step << std::endl;

        if(step < 20)
        {
            if(settings.time_dependent_buondary == true)
            {
                navierstokes.inlet_velocity_boundary.time = time;
            }

            if(settings.algorithm == "linear_imex")
                navierstokes.solve_imex(step);
            else
                navierstokes.solve(step);

            if(settings.time_dependent_buondary == true)
            {
                temperature.temperature_imposed.time = time;
            }

            temperature.solve_linear_ns(navierstokes);
        }


        statistics->enter_subsection("Generating Particles");
        deallog << "Generating Particles..." << std::endl;
        particles_dynamic.generate_particles();
        statistics->leave_subsection();

        statistics->enter_subsection("Advecting Particles");
        deallog << "Advecting Particles..." << std::endl;
        particles_dynamic.advect_particles(step);
        statistics->leave_subsection();




        csv_info << time << "," << navierstokes.l2_norm_div_u << std::endl;

        if (step % settings.output == 0) {
            statistics->enter_subsection("Output Results");

            particles_dynamic.output_results(step);

            if(step < 20)
            {
                DataOut<dim> data_out;
                std::ofstream output(
                        settings.output_directory + settings.output_filename + std::to_string(step) + ".vtk");
                temperature.output_results(data_out);
                navierstokes.output_results(data_out);
                laser.output_results(data_out);
                data_out.write_vtk(output);
            }

            statistics->leave_subsection();
        }
        deallog << std::endl;
        step++;
    }
}

void newton(const Settings &settings, Triangulation<dim> &triangulation)
{
    std::ofstream csv_info(settings.output_directory + "newton_navierstokes.csv");

    NavierStokesNewton navierstokes(settings, triangulation);

    Temperature temperature(settings, triangulation);
    temperature.setup_system();

    deallog << "Number of velocity degrees of freedom: "
            << navierstokes.dofs_per_block[0] << std::endl
            << "Number of pressure degrees of freedom: "
            << navierstokes.dofs_per_block[1] << std::endl
            << "Number of temperature degrees of freedom: "
            << temperature.dof_handler_temperature.n_dofs() << std::endl;

    deallog << std::endl;

    //ParticlesDynamic particles_dynamic(settings,navierstokes,temperature,triangulation);

    statistics->leave_subsection();

    navierstokes.update_boundary = false;

    if (settings.reynolds_number > 1000.0)
    {
        std::cout << "Searching for initial guess ..." << std::endl;
        const double step_size = 2000.0;
        navierstokes.compute_initial_guess(false, step_size);
        std::cout << "Found initial guess." << std::endl;
        std::cout << "Computing solution with target Re = " << settings.reynolds_number << std::endl;
        navierstokes.viscosity = 1.0 / settings.reynolds_number;
    }
    else
    {
        navierstokes.compute_initial_guess(true);
    }

    int n_step = static_cast<int>(settings.simulation_time / settings.time_step);
    int step = 1;
    for (double time = settings.time_step; time <= settings.simulation_time; time += settings.time_step) {
        deallog << "Time: " << time << " step/tot: " << step << "/" << n_step << std::endl;

        if(settings.algorithm == "newton")
        {
            if(settings.time_dependent_buondary == true)
            {
                navierstokes.update_boundary_values(time);
            }
            navierstokes.direct_newton_iteration(settings.newton_tolerance, settings.max_newton_iterations);
        }

        if(settings.algorithm == "quasi_newton")
        {
            if(settings.time_dependent_buondary == true)
            {
                navierstokes.update_boundary_values(time);
            }
            navierstokes.direct_quasi_newton_iteration(settings.newton_tolerance, settings.max_newton_iterations);
        }

        if(settings.time_dependent_buondary == true)
        {
            temperature.temperature_imposed.time = time;
        }

        temperature.solve_nonlinear_ns(navierstokes);

        csv_info << time << "," << navierstokes.l2_norm_div_u << std::endl;

        if (step % settings.output == 0)
        {
            statistics->enter_subsection("Output Results");

            DataOut<dim> data_out;
            std::ofstream output(
                    settings.output_directory + settings.output_filename + std::to_string(step) + ".vtk");
            temperature.output_results(data_out);
            navierstokes.output_results(data_out);
            data_out.write_vtk(output);

            statistics->leave_subsection();
        }

        deallog << std::endl;
        step++;
    }
}

